@@ -17,7 +17,7 @@ BEGIN {
     use Exporter ();
     our ( $VERSION, @EXPORT_OK, %EXPORT_TAGS );
 
-    $VERSION = '2.3.0';
+    $VERSION = '2.6.0';
 
     use base qw(Exporter);
     %EXPORT_TAGS = ();    # eg: TAG => [ qw!name1 name2! ],
@@ -246,6 +246,7 @@ sub add_defaults {
         window_x                            => undef,
         window_y                            => undef,
         'thumb panel'                       => 100,
+        viewer_tools                        => 100,
         scan_window_width                   => undef,
         scan_window_height                  => undef,
         message_window_width                => 600,
@@ -380,7 +381,7 @@ sub add_defaults {
             t                   => 1,
         },
         'scan-reload-triggers' => qw(mode),
-        message                => undef,
+        message                => {},
     );
     if (
         defined $SETTING->{frontend}
