@@ -24,10 +24,11 @@ Readonly my $MAXVAL_8_BIT   => 2**$_8_BIT - 1;
 Readonly my $_16_BIT        => 16;
 Readonly my $MAXVAL_16_BIT  => 2**$_16_BIT - 1;
 Readonly my $LARGE_STATUS   => 99;
+Readonly my $NOT_FOUND      => -1;
 my $uuid_object = Data::UUID->new;
 my $EMPTY       = q{};
 
-our $VERSION = '2.3.0';
+our $VERSION = '2.6.0';
 
 my ( $prog_name, $logger, %callback, $_self );
 
@@ -142,6 +143,24 @@ sub open_device {
     return;
 }
 
+sub close_device {
+    my ( $class, %options ) = @_;
+
+    my $uuid = $uuid_object->create_str;
+    $callback{$uuid}{started}  = $options{started_callback};
+    $callback{$uuid}{running}  = $options{running_callback};
+    $callback{$uuid}{finished} = sub {
+        $_self->{device_name} = $options{device_name};
+        $options{finished_callback}->();
+    };
+    $callback{$uuid}{error} = $options{error_callback};
+    my $sentinel =
+      _enqueue_request( 'close',
+        { uuid => $uuid, device_name => $options{device_name} } );
+    _monitor_process( $sentinel, $uuid );
+    return;
+}
+
 sub find_scan_options {
     my (
         $class,             $started_callback, $running_callback,
@@ -211,6 +230,7 @@ sub scan_page_finished_callback {
             and $status != SANE_STATUS_EOF )
       )
     {
+        if ( $_self->{abort_scan} ) { unlink $path }
         _enqueue_request( 'cancel', { uuid => $uuid_object->create_str } );
         if ( _scanned_enough_pages( $status, $options{npages}, $n_scanned ) ) {
             if ( defined $options{finished_callback} ) {
@@ -362,8 +382,14 @@ sub check_return_queue {
                 $callback{ $data->{uuid} }{started}->();
             }
             if ( defined $callback{ $data->{uuid} }{finished} ) {
-                $callback{ $data->{uuid} }{finished}
-                  ->( _thaw_deref( $data->{info} ), $data->{status} );
+                if ( $data->{process} eq 'set-option' ) {
+                    $callback{ $data->{uuid} }{finished}
+                      ->( $data->{info}, $data->{status} );
+                }
+                else {
+                    $callback{ $data->{uuid} }{finished}
+                      ->( _thaw_deref( $data->{info} ), $data->{status} );
+                }
                 delete $callback{ $data->{uuid} };
             }
         }
@@ -371,6 +397,39 @@ sub check_return_queue {
     return Glib::SOURCE_CONTINUE;
 }
 
+sub _log2 {
+    my ($n) = @_;
+    return log($n) / log 2;
+}
+
+sub decode_info {
+    my ($info) = @_;
+    if ( $info == 0 ) { return 'none' }
+    my @opts =
+      qw(SANE_INFO_INEXACT SANE_INFO_RELOAD_OPTIONS SANE_INFO_RELOAD_PARAMS);
+    my @this;
+    my $n = _log2($info);
+    if ( $n > int $n ) {
+        $n = int($n) + 1;
+    }
+    my $i = @opts;
+    while ( $n > $i ) {
+        if ( $info >= 2**( $n - 1 ) ) {
+            push @this, q{?};
+            $info -= 2**( $n - 1 );
+        }
+        --$n;
+    }
+    while ( $n > $NOT_FOUND ) {
+        if ( $info >= 2**$n ) {
+            push @this, $opts[$n];
+            $info -= 2**$n;
+        }
+        --$n;
+    }
+    return join ' + ', @this;
+}
+
 sub _thread_main {
     my ($self) = @_;
 
@@ -388,6 +447,16 @@ sub _thread_main {
                 _thread_open_device( $self, $request->{uuid},
                     $request->{device_name} )
             }
+            when ('close') {
+                if ( defined( $self->{device_handle} ) ) {
+                    $logger->debug("closing device '$self->{device_name}'");
+                    undef $self->{device_handle};
+                }
+                else {
+                    $logger->debug(
+                        'Ignoring close_device() call - no device open.');
+                }
+            }
             when ('get-options') {
                 _thread_get_options( $self, $request->{uuid} )
             }
@@ -463,7 +532,8 @@ sub _thread_open_device {
     my $status = SANE_STATUS_GOOD;
     try {
         $self->{device_handle} = Image::Sane::Device->open($device_name);
-        $logger->debug("opened device '$device_name'");
+        $self->{device_name}   = $device_name;
+        $logger->debug("opened device '$self->{device_name}'");
     }
     catch {
         $status = $_->status;
@@ -569,28 +639,29 @@ sub _thread_set_option {
         $status = $_->status;
     };
     if ( $logger->is_info ) {
-        $logger->info( "sane_set_option $index ($opt->{name})"
+        $logger->info(
+                "sane_set_option $index ($opt->{name})"
               . ( $opt->{type} == SANE_TYPE_BUTTON ? $EMPTY : " to $value" )
               . " returned status $status ("
               . Image::Sane::strstatus($status)
               . ') with info '
-              . ( defined $info ? $info : 'undefined' ) );
-    }
-
-    # $info could be undefined if status is invalid
-    if ( defined $info and $info & SANE_INFO_RELOAD_OPTIONS ) {
-        $status = _thread_get_options( $self, $uuid );
-    }
-    else {
-        $self->{return}->enqueue(
-            {
-                type    => 'finished',
-                process => 'set-option',
-                uuid    => $uuid,
-                status  => $status,
-            }
+              . (
+                defined $info
+                ? sprintf( '%d (%s)',
+                    $info, Gscan2pdf::Frontend::Image_Sane::decode_info($info) )
+                : 'undefined'
+              )
         );
     }
+    $self->{return}->enqueue(
+        {
+            type    => 'finished',
+            process => 'set-option',
+            uuid    => $uuid,
+            status  => $status,
+            info    => $info,
+        }
+    );
     return;
 }
 
@@ -739,6 +810,7 @@ sub _thread_scan_page {
         $status = $_->status;
         _thread_throw_error( $self, $uuid, 'scan-page', $status,
             "$prog_name: sane_start: " . $_->error );
+        unlink $path;
     };
     if ( $status != SANE_STATUS_GOOD ) { return }
 
